"""
Allele
"""

import math

import gadapt.adapters.string_operation.ga_strings as ga_strings
import gadapt.ga_model.definitions as definitions
from gadapt.ga_model.gene import Gene
from gadapt.ga_model.ranking_model import RankingModel


class Allele(RankingModel):
    def __init__(self, gen_variable, var_value=None):
        """
        Allele class. Allele is a part of chromosome.
        It contains concrete value of Gene.
        Args:
            gen_variable: Gene which defines the allele
            var_value: Variable value
        """
        super().__init__()
        self.gene = gen_variable
        self.variable_value = var_value
        self._rank = -1
        self._cummulative_probability = definitions.FLOAT_NAN
        if self.variable_value is None or math.isnan(self.variable_value):
            self.set_random_value()

    def __str__(self) -> str:
        return self._to_string()

    def _to_string(self):
        return ga_strings.gene_value_to_string(self)

    @property
    def gene(self) -> Gene:
        """
        Gene which defines the allele
        """
        return self._gene

    @gene.setter
    def gene(self, value: Gene):
        if not isinstance(value, Gene):
            raise
        self._gene = value

    @property
    def variable_value(self):
        """
        Variable value
        """
        return self._variable_value

    @variable_value.setter
    def variable_value(self, value):
        self._variable_value = value

    def set_random_value(self):
        """
        Sets a random value for the variable_value property
        """
        self.variable_value = self.gene.make_random_value()



from gadapt.ga_model.chromosome import Chromosome
from gadapt.ga_model.allele import Allele
from gadapt.utils import ga_utils
from operations.chromosome_update.base_chromosome_updater import (
    BaseChromosomeUpdater,
)


class ParentDiversityChromosomeUpdater(BaseChromosomeUpdater):
    """
    Updates chromosome for the parent diversity purpose.
    """

    def __init__(self):
        self._genetic_diversity = None

    def _get_genetic_diversity(self, mother_gene, father_gene) -> float:
        return abs(mother_gene.variable_value - father_gene.variable_value) / (
            father_gene.gene.max_value - father_gene.gene.min_value
        )

    def _get_parent_diversity(self):
        return round(ga_utils.average(self._genetic_diversity), 2)

    def chromosome_prepare_update(self, mother_gene: Allele, father_gene: Allele):
        if mother_gene is None or father_gene is None:
            return
        self._genetic_diversity.append(
            self._get_genetic_diversity(mother_gene, father_gene)
        )

    def chromosome_update(self, offspring1: Chromosome, offspring2: Chromosome):
        parent_diversity = self._get_parent_diversity()
        offspring1.parent_diversity_coefficient = parent_diversity
        offspring2.parent_diversity_coefficient = parent_diversity

    def chromosome_start_update(self, *args, **kwargs):
        self._genetic_diversity = []



from gadapt.ga_model.chromosome import Chromosome
from gadapt.ga_model.allele import Allele


class BaseChromosomeUpdater:
    """
    Updates chromosome
    """

    def chromosome_prepare_update(self, mother_gene: Allele, father_gene: Allele):
        pass

    def chromosome_update(self, offspring1: Chromosome, offspring2: Chromosome):
        pass

    def chromosome_start_update(self, *args, **kwargs):
        pass



import math

from adapters.ga_logging.logging_settings import gadapt_log_error
from gadapt.operations.mutation.population_mutation.base_chromosome_mutation_rate_determinator import (
    BaseChromosomeMutationRateDeterminator,
)


class CostDiversityChromosomeMutationRateDeterminator(
    BaseChromosomeMutationRateDeterminator
):
    """
    Determines the number of chromosomes to be mutated in a population based on the cost diversity of the population.
    """

    def __init__(
        self,
    ) -> None:
        super().__init__()

    def _get_cost_diversity_coefficient(self):
        cost_diversity_coefficient = float(
            self.population.absolute_cost_diversity
            / self.population.absolute_cost_diversity_in_first_population
        )
        if cost_diversity_coefficient > 1.0:
            cost_diversity_coefficient = 1.0
        return cost_diversity_coefficient

    def _get_mutation_rate(self) -> float:
        if (
            self.population.absolute_cost_diversity_in_first_population is None
            or math.isnan(self.population.absolute_cost_diversity_in_first_population)
            or self.population.absolute_cost_diversity is None
            or math.isnan(self.population.absolute_cost_diversity)
        ):
            gadapt_log_error("absolute_cost_diversity not set!")
            return 1.0
        return 1.0 - self._get_cost_diversity_coefficient()

    def _get_number_of_mutation_chromosomes(self) -> int:

        mutation_rate = self._get_mutation_rate()
        f_return_value = mutation_rate * float(self.max_number_of_mutation_chromosomes)
        return round(f_return_value)



from math import isnan

import gadapt.utils.ga_utils as ga_utils
from adapters.ga_logging.logging_settings import gadapt_log_error
from gadapt.operations.mutation.chromosome_mutation.random_gene_mutation_rate_determinator import (
    RandomGeneMutationRateDeterminator,
)


class CrossDiversityGeneMutationRateDeterminator(RandomGeneMutationRateDeterminator):
    """
    Determines the number of mutation alleles in a chromosome based on the cross diversity coefficient of the gene.
    """

    def __init__(
        self,
    ) -> None:
        super().__init__()

    def _get_mutation_rate(self, genes) -> float:
        avg_rsd = ga_utils.average([g.cross_diversity_coefficient for g in genes])
        if avg_rsd > 1:
            avg_rsd = 1
        if avg_rsd < 0:
            avg_rsd = 0
        return avg_rsd

    def _get_number_of_mutation_genes(self) -> int:
        genes = [g.gene for g in self.chromosome]
        if any(
            g.cross_diversity_coefficient is None
            or isnan(g.cross_diversity_coefficient)
            for g in genes
        ):
            gadapt_log_error("cross_diversity_coefficient not set!")
            return super()._get_number_of_mutation_genes()

        mutation_rate = self._get_mutation_rate(genes)
        limit_number_of_mutation_genes = mutation_rate * float(
            self.max_number_of_mutation_genes
        )
        limit_number_of_mutation_genes_rounded = round(limit_number_of_mutation_genes)
        if limit_number_of_mutation_genes_rounded == 0:
            limit_number_of_mutation_genes_rounded = 1
        self.max_number_of_mutation_genes = limit_number_of_mutation_genes_rounded
        return super()._get_number_of_mutation_genes()



class BaseGeneUpdater:
    """
    Base class for variable update
    """

    def __init__(self):
        self.population = None

    def update_genes(self, population):
        self.population = population
        self._update_genes()

    def _update_genes(self):
        pass
